------------------------------------------------------------------------
-- Example program demonstrating ESF Arrays
-- John T. O'Donnell
-- April 2013
------------------------------------------------------------------------

{- For documentation of the ESFA system, see the user's guide in the
doc directory.  There is a source text file in markdown format
intended for pandoc, and there may be an html version.

There are several programs that demonstrate sequences of ESFA
operations.  Each of these has type IO ().  To run an example,
  1. (shell prompt)  cabal install
  2. (shell prompt)  ghci examples/EsfaExamples
  3. (ghci prompt)   runPARLE  (or any other example, or main)

runprog1
runprog2
runprog3
runprog4
runprog5

runPARLE
   the example from the paper in PARLE'93
runtree_parle
   the example from the paper in PARLE'93   (redundant?)

run_single_threaded_array
   single threaded updates
runUpdateMulti
   multi-threaded updates
run_multiple_single_threaded_arrays
run_multiple_single_threaded_arrays_interleaved
runprogSeqDeletes
   a sequence of deletes; watch the effect of sharing
runFigSimple
   example used in Fig 1(?) in submitted paper

-}

------------------------------------------------------------------------
module EsfaExample where

import Control.Monad.IO.Class
import Control.Monad.Trans.State.Strict
import Data.Array.ESFA.Spec
import Data.Array.ESFA.SimArch

main :: IO ()
main =
  do dashline
     runprog1
     dashline
     runprog2
     dashline
     runprog3
     dashline

------------------------------------------------------------------------
-- Types for esfa element values

-- We need a single data type that will be used for all array
-- elements.  This could be something like Int, but if we want to have
-- several different types in esf arrays, it's necessary to define a
-- single algebraic data type for all of them, like this:

data Value
  = Ival Int
  | Dval Double
  | Sval String
  | Tval Tree
  deriving (Read, Eq)

instance Show Value where
  show (Ival i) = show i
  show (Dval x) = show x
  show (Sval s) = s
  show (Tval t) = show t

data Tree
  = Leaf Int
  | Node Int Tree Tree
  deriving (Read, Show, Eq)

-- Because of this tagging, you can in effect have values of different
-- types in an esf array.

-- When you do an update you need to tell the esf system whether the
-- value you're updating with is one of yours (Value above), or an esf
-- array name.  This is necessary because the esf system needs to know
-- if it contains an array name in an array element.  If you have a
-- user value x, then use EsfUserOrdVal x as the argument to update, but
-- if it's an esf array name, use EsfAVal Aname.

-- Example: let's update an esf array with a Double
--    Just anew <- esfmUpdate "anew" olda (100, EsfUserOrdVal (Dval 3.14))
-- Now let's store anew (which is an esf array name) into another array:
--    Just new2 <- esfmUpdate "new2" anew (101, EsfAVal anew)

-- It may be convenient to define some helper functions for the data types
-- you want to put into esf arrays, like this:

mkInt :: Int -> EsfEltVal Value
mkInt i = EsfUserOrdVal (Ival i)

mkDbl :: Double -> EsfEltVal Value
mkDbl x = EsfUserOrdVal (Dval x)

mkStr :: String -> EsfEltVal Value
mkStr s = EsfUserOrdVal (Sval s)

mkTr :: Tree -> EsfEltVal Value
mkTr t = EsfUserRefVal (Tval t)

mkArr :: AName -> EsfEltVal Value
mkArr a = EsfAVal a

------------------------------------------------------------------------

-- Utilities

printStats :: Show a => StateT (PtcRunState u a) IO ()
printStats =
  do stats <- esfmAnalyzeMachineState
     liftIO (putStrLn (show stats))

-- Remove zombies

cleanup :: Show a => StateT (PtcRunState u a) IO ()
cleanup =
  do liftIO (putStrLn "Starting zombie cleanup")
     n <- esfmKillZombies f
     liftIO (putStrLn ("Zapped " ++ show n ++ " zombies"))
  where f z = liftIO (putStrLn ("  decr ref count: " ++ show z))

-- Print the arrays, statistics, and machine state

printFullStatus =
  do liftIO $ putStrLn "\n***** system status start *****"
     esfmPrintSpecArrays
     esfmPrintMachineArrays
     liftIO $ putStrLn "ptcPrintState"
     ptcPrintState
     stats <- esfmAnalyzeMachineState
     liftIO (putStrLn (show stats))
     liftIO $ putStrLn "***** system status end *****\n"

dashline = putStrLn (take 72 (repeat '-'))

------------------------------------------------------------------------
-- prog1 illustrates basic updates and lookups

runprog1 = runtree 5 () prog1

prog1 :: StESFA () Value ()
prog1 =
  do liftIO $ putStrLn "\nExample program for ESFA: prog1"

-- You first have to initialize the system
     esfmInitialize

-- Now do some esfa operations
     Just a1 <- esfmUpdate "a1" esfmEmpty (100, mkInt 101)
     Just a2 <- esfmUpdate "a2" a1 (200, mkInt 201)
     Just a3 <- esfmUpdate "a2" a2 (300, mkInt 301)
     Just x1 <- esfmLookup "a3" a3 100
     Just x2 <- esfmLookup "a3" a3 200
     Just x3 <- esfmLookup "a3" a3 300

-- Print the values so we can see if they're correct
     liftIO $ putStrLn (show x1 ++ " " ++ show x2 ++ " " ++ show x3)

-- Create some more arrays, do a lookup, then delete
     liftIO (putStrLn "Try some deletions")
     esfmPrintState

     Just a4 <- esfmUpdate "a4" esfmEmpty (500, mkStr "aa")
     Just x <- esfmLookup "a4" a4 500
     liftIO (putStrLn ("x = " ++ show x))
     esfmDelete a4

     Just a5 <- esfmUpdate "a5" esfmEmpty  (501, mkStr "bb")
     Just x <- esfmLookup "a4" a5 501
     liftIO (putStrLn ("x = " ++ show x))
     esfmDelete a5

     Just a6 <- esfmUpdate "a6" esfmEmpty  (502, mkStr "cc")
     Just x <- esfmLookup "a4" a6 502
     liftIO (putStrLn ("x = " ++ show x))
     esfmDelete a6

     esfmPrintState

-- Nothing else has to be done, the program can terminate.
-- However, let's print some dump and trace data
     liftIO $ putStrLn "\nThis is the internal esfa machine state:"
     esfmPrintState
     liftIO $ putStrLn "This is the trace state, showing readable arrays:"
     ptcDumpTrace

------------------------------------------------------------------------

-- prog2 illustrates delete and cleanup.  The () parameter is the
-- user's local state; see prog3 for an explanation of the state.

runprog2 =  runtree 5 () prog2

prog2 :: StESFA () Value ()
prog2 =
  do liftIO $ putStrLn "\nExample program for ESFA: prog2"
     esfmInitialize
     esfmSetOption EsfmTraceOperation
     Just b1 <- esfmUpdate "b1" esfmEmpty (1, mkInt 2)
     Just a1 <- esfmUpdate "a1" esfmEmpty (100, mkTr (Leaf 100))
     Just a2 <- esfmUpdate "a2" a1 (101, mkTr (Leaf 101))
     Just a3 <- esfmUpdate "a3" a2 (102, mkTr (Node 3 (Leaf 4) (Leaf 5)))
     esfmDelete a3
     esfmDelete a2
     esfmDelete a1
     cleanup
     esfmPrintState

------------------------------------------------------------------------
-- prog3: Working with the user's state

-- 5 is the machine size (2^5 cells)
-- mlsInit is the initial value of my user state
-- prog3 is the algorithm

runprog3 = runtree 5 mlsInit prog3

-- The user's local state can be any type, and the esf system won't do
-- anything with it.  It could be an Int, but it's often useful to
-- have several fields.  In this example, we have an Int called mlsA
-- and a string called mlsStr

data MyLocalState = MyLocalState
  { mlsA :: Int
  , mlsStr :: String
  } deriving Show

-- It's necessary to provide an initial value for the state
mlsInit :: MyLocalState
mlsInit = MyLocalState { mlsA = 13, mlsStr = "Hello" }

-- In the type of the script, the type constructor is EsfmRun.  Its
-- first parameter is the type of the users's private local state,
-- while the second parameter is the type of array element that the
-- program will use.  prog3 doesn't do much with esfa operations; its
-- purpose is to show how to use a state with get and put in a program
-- doing esfa operations.  The example shows some direct usage of get
-- and put, and below are defined some typical helper functions that
-- operate on the state.

-- Strictly speaking, this may not be the best way to handle the
-- state, because it makes the esfa system state visible to the user.
-- That might be useful for experimental purposes, but if the user
-- messes up the esfa system state, the system won't work.  Note that
-- the code below always puts back the sysstate unchanged, when a put
-- is used.  Later it might be useful to use two layers of StateT to
-- hide the system state, but for now it's visible to the user, which
-- may be useful (for example, it would be possible to have an
-- experimental program that grabs the esfa system state and gathers
-- some statistics on it, without needing to modify the ESFA module.

prog3 :: StESFA MyLocalState Value ()
prog3 =
  do liftIO $ putStrLn "\nExample program for ESFA: prog3"
     esfmInitialize

     (mystate,sysstate) <- get
     liftIO $ putStrLn ("mystate = " ++ show mystate)
     let mystate' = mystate { mlsA = (mlsA mystate) + 1 }
     put (mystate',sysstate)
     liftIO $ putStrLn ("mystate = " ++ show mystate)

     esfmSetOption EsfmTraceOperation
     Just b1 <- esfmUpdate "b1" esfmEmpty (1, mkInt 2)
     Just b2 <- esfmUpdate "b2" b1 (2, mkTr (Leaf 100))
     x <- esfmLookup "b2" b2 1
     y <- esfmLookup "b2" b2 2
     liftIO $ putStrLn ("Results are " ++ show x ++ " and " ++ show y)

     (mystate,sysstate) <- get
     let mystate' = mystate { mlsStr = "Bye" }
     put (mystate',sysstate)

     liftIO $ putStrLn "try the mystate helper functions"
     prMyState
     incrMyInt
     setMyString "does it work?"
     prMyState

prMyState =
  do (mystate,sysstate) <- get
     liftIO $ putStrLn ("mystate = " ++ show mystate)

incrMyInt =
  do (mystate,sysstate) <- get
     let mystate' = mystate {mlsA = mlsA mystate + 1}
     put (mystate',sysstate)

setMyString xs =
  do (mystate,sysstate) <- get
     let mystate' = mystate {mlsStr = xs}
     put (mystate',sysstate)

------------------------------------------------------------------------
-- prog4: example of repeat (update - delete)

runprog4 = runtree 5 () prog4

prog4 :: StESFA () Value ()
prog4 =
  do liftIO $ putStrLn "\nExample prog4: repeat (update - delete)g4"
     esfmInitialize
     esfmSetOption EsfmTraceOperation
     Just a1 <- esfmUpdate "a1" esfmEmpty (1, mkInt 11)
     Just a2 <- esfmUpdate "a2" a1 (2, mkInt 22)
     Just b1 <- esfmUpdate "b1" esfmEmpty (100, mkInt 101)
     Just c1 <- esfmUpdate "c1" esfmEmpty (5, mkInt 55)
     liftIO $ putStrLn "created a1,a2,b1,c1"
     esfmPrintState

     liftIO $ putStrLn "-------------------"
     liftIO $ putStrLn "Existing array names:"
     xs <- esfmGetArrayNames
     liftIO $ putStrLn (show xs)

     esfmPrintSpecArrays
     esfmPrintMachineArrays
     stats <- esfmAnalyzeMachineState
     liftIO (putStrLn (show stats))

     liftIO $ putStrLn "-------------------"
     Just b2 <- esfmUpdate "b2" b1 (200, mkInt 201)
     liftIO $ putStrLn ("b2 = " ++ show b2)
     esfmPrintState
     liftIO $ putStrLn ("deleting b1 = " ++ show b1)
     esfmDelete b1
     esfmPrintState

     liftIO $ putStrLn "-------------------"
     Just b3 <- esfmUpdate "b3" b2 (300, mkInt 301)
     liftIO $ putStrLn ("b3 = " ++ show b3)
     esfmPrintState
     esfmDelete b2
     esfmPrintState

     liftIO $ putStrLn "-------------------"
     Just b4 <- esfmUpdate "b4" b3 (400, mkInt 401)
     liftIO $ putStrLn ("b4 = " ++ show b4)
     esfmPrintState
     esfmDelete b3
     esfmPrintState

     liftIO $ putStrLn "-------------------"
     Just b5 <- esfmUpdate "b5" b4 (500, mkInt 501)
     liftIO $ putStrLn ("b5 = " ++ show b5)
     esfmPrintState
     esfmDelete b4

     esfmPrintState
     esfmPrintSpecArrays
     esfmPrintMachineArrays
     stats <- esfmAnalyzeMachineState
     liftIO (putStrLn (show stats))

------------------------------------------------------------------------
-- prog5: example of repeat (update - delete) without sharing

runprog5 = runtree 5 () prog5

prog5 :: StESFA () Value ()
prog5 =
  do liftIO $ putStrLn "\nExample prog5: repeat (update - delete) no sharing"
     esfmInitialize
     esfmSetOption EsfmTraceOperation
     Just a1 <- esfmUpdate "a1" esfmEmpty (1, mkInt 11)
     Just a2 <- esfmUpdate "a2" a1 (2, mkInt 22)
     Just b1 <- esfmUpdate "b1" esfmEmpty (100, mkInt 101)
     printFullStatus
     esfmDelete b1
     printFullStatus
     Just c1 <- esfmUpdate "c1" a2 (5, mkInt 55)
     printFullStatus
     esfmDelete c1
     printFullStatus
     Just d1 <- esfmUpdate "d1" a2 (6, mkInt 66)
     printFullStatus
     esfmDelete d1
     printFullStatus
     Just e1 <- esfmUpdate "e1" a2 (7, mkInt 77)
     printFullStatus
     esfmDelete e1
     printFullStatus
     Just f1 <- esfmUpdate "f1" a2 (8, mkInt 88)
     printFullStatus
     esfmDelete f1
     printFullStatus
     Just g1 <- esfmUpdate "g1" a2 (9, mkInt 99)
     printFullStatus
     esfmDelete g1
     printFullStatus

------------------------------------------------------------------------

-- progSeqDeletes: string of updates; string of deletes in the order
-- the elements were added does nothing until the last delete

runprogSeqDeletes = runtree 5 () progSeqDeletes

progSeqDeletes :: StESFA () Value ()
progSeqDeletes =
  do liftIO $ putStrLn "\nExample prog6: repeat (update - delete) no sharing"
     let showval (EsfUserOrdVal i) = show i
     esfmInitialize
     esfmSetOption EsfmTraceOperation
     Just a1 <- esfmUpdate "a1" esfmEmpty (1, mkInt 11)
     Just b1 <- esfmUpdate "b1" esfmEmpty (21, mkInt 201)
     Just b2 <- esfmUpdate "b2" b1 (22, mkInt 202)
     Just b3 <- esfmUpdate "b3" b2 (23, mkInt 203)

     liftIO $ putStrLn ("Created some arrays:" ++ show[a1,b1,b2,b3])
     printFullStatus
     latexMachineState showval

     esfmDelete b1
     liftIO $ putStrLn "Just deleted deleted b1"
     printFullStatus
     latexMachineState showval

     esfmDelete b2
     liftIO $ putStrLn "Just deleted deleted b2"
     printFullStatus
     latexMachineState showval

     esfmDelete b3
     liftIO $ putStrLn "Just deleted deleted b3"
     printFullStatus
     latexMachineState showval

{-
     Just c3 <- esfmUpdate "c3" c2 (7, mkInt 77)
     Just c4 <- esfmUpdate "c4" c3 (8, mkInt 88)
     Just c5 <- esfmUpdate "c5" c4 (9, mkInt 99)
     printFullStatus
     latexMachineState showval
     esfmDelete c1
     esfmDelete c2
     esfmDelete c3
     esfmDelete c4

     esfmDelete c5
     liftIO $ putStrLn "Just deleted c5"
     printFullStatus
     latexMachineState showval
-}


------------------------------------------------------------------------
-- progFigSimple: simple string of updates, building one array

runFigSimple = runtree 5 () progFigSimple

progFigSimple :: StESFA () Value ()
progFigSimple =
  do liftIO $ putStrLn "\nExample FigSimple"
     let showval (EsfUserOrdVal i) = show i
     esfmInitialize
     esfmSetOption EsfmTraceOperation
     Just a1 <- esfmUpdate "a1" esfmEmpty (1, mkInt 11)
     printFullStatus
     latexMachineState showval

     Just a2 <- esfmUpdate "a2" a1 (2, mkInt 22)
     printFullStatus
     latexMachineState showval

     Just a3 <- esfmUpdate "a3" a2 (3, mkInt 33)
     printFullStatus
     latexMachineState showval

     Just a4 <- esfmUpdate "a4" a3 (4, mkInt 44)
     printFullStatus
     latexMachineState showval




------------------------------------------------------------------------
-- progParle - example from PARLE'93 paper

runPARLE :: IO ()
runPARLE = runtree 4 () runtree_PARLE

runtree_PARLE =
  do liftIO (putStrLn "test example from PARLE 93")
     let showval (EsfUserOrdVal i) = show i
     esfmInitialize
     Just a1 <- esfmUpdate "a1" esfmEmpty (100, mkInt 101)
     Just a2 <- esfmUpdate "a2" a1 (200, mkInt 201)
     Just a3 <- esfmUpdate "a3" a1 (300, mkInt 301)
     Just a4 <- esfmUpdate "a4" a2 (400, mkInt 401)
     Just a5 <- esfmUpdate "a5" a2 (500, mkInt 501)
     printFullStatus
     latexMachineState showval
     Just a6 <- esfmUpdate "a6" a3 (600, mkInt 601)
     printFullStatus
     latexMachineState showval

-- a1 = 100-101
-- a2 = 200-201 100-101 
-- a3 = 300-301 100-101
-- a4 = 400-401 200-201 100-101 
-- a5 = 500-501 200-201 100-101 
-- a6 = 600-601 300-301 100-101

     x1 <- esfmLookup "a4" a4 (400)
     x2 <- esfmLookup "a5" a5 (500)
     x3 <- esfmLookup "a5" a5 (200)
     x4 <- esfmLookup "a5" a5 (100)
     liftIO (putStrLn (show [x1,x2,x3,x4]))
     esfmPrintState


------------------------------------------------------------------------
-- progUpdateMulti updates with multithreaded access

runUpdateMulti :: IO ()
runUpdateMulti = runtree 4 () runtree_UpdateMulti

runtree_UpdateMulti =
  do liftIO (putStrLn "Update Multi")
     let showval (EsfUserOrdVal i) = show i
     esfmInitialize
     esfmSetOption EsfmTraceOperation

     Just a1 <- esfmUpdate "a1" esfmEmpty (1, mkInt 11)
     Just a2 <- esfmUpdate "a2" esfmEmpty (2, mkInt 22)
     Just a3 <- esfmUpdate "a3" esfmEmpty (3, mkInt 33)

     liftIO separator_line
     liftIO (putStrLn "Have created a1,a2,a3")
     printFullStatus
     latexMachineState showval

     Just a4 <- esfmUpdate "a4" a2 (4, mkInt 44)
     liftIO separator_line
     liftIO (putStrLn "Have created a4")
     printFullStatus
     latexMachineState showval


{-
Crufy


------------------------------------------------------------------------
-- Test data for ptc



myintval :: Int -> EsfEltVal Int
myintval x = EsfUserOrdVal x

runtree_parle :: IO ()
runtree_parle = runtree 3 () runtree_test

runtree_test =
  do liftIO (putStrLn "test example from PARLE 93")
     esfmInitialize
     Just a1 <- ptcUpdate "a1" 0 (100, myintval 101)
     Just a2 <- ptcUpdate "a2" a1 (200, myintval 201)
     Just a3 <- ptcUpdate "a3" a1 (300, myintval 301)
     Just a4 <- ptcUpdate "a4" a2 (400, myintval 401)
     Just a5 <- ptcUpdate "a5" a2 (500, myintval 501)
     Just a6 <- ptcUpdate "a6" a3 (600, myintval 601)

-- a1 = 100-101
-- a2 = 200-201 100-101 
-- a3 = 300-301 100-101
-- a4 = 400-401 200-201 100-101 
-- a5 = 500-501 200-201 100-101 
-- a6 = 600-601 300-301 100-101

     x1 <- ptcLookup "a4" a4 (400)
     x2 <- ptcLookup "a5" a5 (500)
     x3 <- ptcLookup "a5" a5 (200)
     x4 <- ptcLookup "a5" a5 (100)

     ptcPrintState

run_single_threaded_array =
  do liftIO (putStrLn "example: single threaded array")
     esfmInitialize
     Just a0 <- ptcUpdate "a0" 0 (100, myintval 10)
     Just a1 <- ptcUpdate "a1" a0 (101, myintval 11)
     Just a2 <- ptcUpdate "a2" a1 (102, myintval 12)
     Just a3 <- ptcUpdate "a3" a2 (103, myintval 13)
     Just a4 <- ptcUpdate "a4" a3 (104, myintval 14)

-- a0 = 100-10
-- a1 = 101-11 100-10
-- a2 = 102-12 101-11 100-10
-- a3 = 103-13 102-12 101-11 100-10
-- a4 = 104-14 103-13 102-12 101-11 100-10

     ptcPrintState

run_multiple_single_threaded_arrays =
  do liftIO (putStrLn "multiple single threaded arrays")
     esfmInitialize
     Just a0 <- ptcUpdate "a0" 0 (100, myintval 10)
     Just a1 <- ptcUpdate "a1" a0 (101, myintval 11)
     Just a2 <- ptcUpdate "a2" a1 (102, myintval 12)
     Just a3 <- ptcUpdate "a3" a2 (103, myintval 13)
     Just a4 <- ptcUpdate "a4" a3 (104, myintval 14)
     Just b0 <- ptcUpdate "b0" 0 (200, myintval 20)
     Just b1 <- ptcUpdate "b1" b0 (201, myintval 21)
     Just b2 <- ptcUpdate "b2" b1 (202, myintval 22)
     Just b3 <- ptcUpdate "b3" b2 (203, myintval 23)
     Just c0 <- ptcUpdate "c0" 0 (300, myintval 30)
     Just c1 <- ptcUpdate "c1" c0 (301, myintval 31)
     Just c2 <- ptcUpdate "c2" c1 (302, myintval 32)
     Just c3 <- ptcUpdate "c3" c2 (303, myintval 33)

-- a0 = 100-10
-- a1 = 101-11 100-10
-- a2 = 102-12 101-11 100-10
-- a3 = 103-13 102-12 101-11 100-10
-- a4 = 104-14 103-13 102-12 101-11 100-10
-- b0 = 200-20
-- b1 = 201-21 200-20
-- b2 = 202-22 201-21 200-20
-- b3 = 203-23 202-22 201-21 200-20
-- c0 = 300-30
-- c1 = 301-31 300-30
-- c2 = 302-32 301-31 300-30
-- c3 = 303-33 302-32 301-31 300-30

     ptcPrintState

run_multiple_single_threaded_arrays_interleaved =
  do liftIO (putStrLn "multiple single threaded arrays interleaved")
     esfmInitialize

     Just a0 <- ptcUpdate "a0" 0 (100, myintval 10)
     Just a1 <- ptcUpdate "a1" a0 (101, myintval 11)
     Just b0 <- ptcUpdate "b0" 0 (200, myintval 20)
     Just a2 <- ptcUpdate "a2" a1 (102, myintval 12)
     Just c0 <- ptcUpdate "c0" 0 (300, myintval 30)
     Just c1 <- ptcUpdate "c1" c0 (301, myintval 31)
     Just a3 <- ptcUpdate "a3" a2 (103, myintval 13)
     Just b1 <- ptcUpdate "b1" b0 (201, myintval 21)
     Just b2 <- ptcUpdate "b2" b1 (202, myintval 22)
     Just a4 <- ptcUpdate "a4" a3 (104, myintval 14)
     Just c2 <- ptcUpdate "c2" c1 (302, myintval 32)
     Just b3 <- ptcUpdate "b3" b2 (203, myintval 23)
     Just c3 <- ptcUpdate "c3" c2 (303, myintval 33)

     ptcPrintState

-}

{- Bit-rotted code that needs revision

------------------------------------------------------------------------
-- test1: empty array contains no elements

{- test1 prints empty[5], empty[6], ..., empty[8].  It's using the dp1
implementation of esfa, with the memory in the initial state
dp1init. -}

test1 =
  do putStrLn "test1"
     printSegment dp1lookup dp1init dp1empty 5 8

------------------------------------------------------------------------
-- test2: Some updates with sharing, and evey update has a different
-- index

{-
a1 = update empty 3 30
a2 = update a1 5 50
a3 = update a1 4 40
a4 = update a2 6 60
a5 = update a2 7 70
a6 = update a3 8 80
-}

test2 =
  do putStrLn "test2"

     putStrLn "\na1 = update empty 3 30"
     let (s1,a1) = upd s0 empty 3 30            -- a1 = update empty 3 30
     printSegment dp1lookup s1 a1 3 8

     putStrLn "\na2 = update a1 5 50"
     let (s2,a2) = upd s1 a1 5 50               -- a2 = update a1 5 50
     printSegment dp1lookup s2 a2 3 8

     putStrLn "\na3 = update a1 4 40"
     let (s3,a3) = upd s2 a1 4 40               -- a3 = update a1 4 40
     printSegment dp1lookup s3 a3 3 8

     putStrLn "\na4 = update a2 6 60"
     let (s4,a4) = upd s3 a2 6 60               -- a4 = update a2 6 60
     printSegment dp1lookup s4 a4 3 8

     putStrLn "\na5 = update a2 7 70"
     let (s5,a5) = upd s4 a2 7 70               -- a5 = update a2 7 70
     printSegment dp1lookup s5 a5 3 8

     putStrLn "\na6 = update a3 8 80"
     let (s6,a6) = upd s5 a3 8 80               -- a6 = update a3 8 80
     printSegment dp1lookup s6 a6 3 8

     putStrLn "\nAt end, in final state, all arrays: a1, a2, ..., a6"
     let finalst = s6
     printSegment dp1lookup finalst a1 3 8
     printSegment dp1lookup finalst a2 3 8
     printSegment dp1lookup finalst a3 3 8
     printSegment dp1lookup finalst a4 3 8
     printSegment dp1lookup finalst a5 3 8
     printSegment dp1lookup finalst a6 3 8

------------------------------------------------------------------------

-- test3: This is the same as test2, except that some updates have an
-- index that gives the new array a different value at some index than
-- the original array.  This is useful for checking that (1) the new
-- array has the right value, and also (2) the old array has not been
-- modified.

{-
a1 = update empty 3 30
a2 = update a1 5 50
a3 = update a1 3 300
a4 = update a2 5 500
a5 = update a2 7 70
a6 = update a3 8 80
-}

test3 =
  do putStrLn "test3"

     putStrLn "\na1 = update empty 3 30"
     putStrLn "expect a1 = 3->30"
     let (s1,a1) = upd s0 empty 3 30            -- a1 = update empty 3 30
     printSegment dp1lookup s1 a1 3 8

     putStrLn "\na2 = update a1 5 50"
     putStrLn "expect a2 = 3->30 5->50"
     let (s2,a2) = upd s1 a1 5 50               -- a2 = update a1 5 50
     printSegment dp1lookup s2 a2 3 8

     putStrLn "\na3 = update a1 3 300"
     putStrLn "expect a3 = 3->300"
     let (s3,a3) = upd s2 a1 3 300               -- a3 = update a1 3 300
     printSegment dp1lookup s3 a3 3 8

     putStrLn "\na4 = update a2 5 500"
     putStrLn "expect a4 = 3->30 5->500"
     let (s4,a4) = upd s3 a2 5 500               -- a4 = update a2 5 500
     printSegment dp1lookup s4 a4 3 8

     putStrLn "\na5 = update a2 7 70"
     putStrLn "expect a5 = 3->30 5->50 7->70"
     let (s5,a5) = upd s4 a2 7 70               -- a5 = update a2 7 70
     printSegment dp1lookup s5 a5 3 8

     putStrLn "\na6 = update a3 8 80"
     putStrLn "expect a6 =  3->300 8->80"
     let (s6,a6) = upd s5 a3 8 80               -- a6 = update a3 8 80
     printSegment dp1lookup s6 a6 3 8

     putStrLn "\nAt end, in final state, all arrays: a1, a2, ..., a6"
     let finalst = s6
     printSegment dp1lookup finalst a1 3 8
     printSegment dp1lookup finalst a2 3 8
     printSegment dp1lookup finalst a3 3 8
     printSegment dp1lookup finalst a4 3 8
     printSegment dp1lookup finalst a5 3 8
     printSegment dp1lookup finalst a6 3 8

------------------------------------------------------------------------

-}
