#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "esftypes.h"

/* Parameters that the user may change */
#define ESF_NCELLS  65536 /* size of machine: number of leaf cells */
#define ESF_TRACE       0 /* trace data is printed iff ESF_TRACE is 1 */

/* ----------------------------------------------------------------------
Summary of the most important user-level functions

ESF_EMPTY
  The constant empty array with no elements.  It always exists,
  and isn't created by the user program.

int esfc_update (int a, int i, int v)
  a2 = esfc_update (a1,i,v)
  a2 is a new array with the same contents as a1, except that a2[i] = v
     (it doesn't matter whether a1[i] was defined)
  The old array a1 is not changed by the update.

int esfc_lookup (int a, int i)
  returns a[i]

int esfc_maxindex (int a)
  returns the maximum index where a has a defined value

int esfc_minindex (int a)
  returns the minimum index where a has a defined value

int esfc_nextindex (int a, int ind)
  returns the next index > ind where a is defined
  If no such index exists, returns INT_MAX

int esfc_previndex (int a, int ind)
  returns the next index < ind where a is defined
  If no such index exists, returns INT_MIN

---------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
			    Machine state
---------------------------------------------------------------------- */

/* The struct cell defines the fields in each leaf cell in the tree
   machine.  This is the cell state (the type s that appears in the
   type of sweep).  The tree nodes don't have a persistent state.

   Name/code mapping.  Keep mapping for (n,c) in cell n.  If there is
   an array with name n, then this cell contains (n,c), and if there
   is no array with name n,, it contains (-n, _). */

struct cell {
  int cellid;             /* k bit word */
  int eltfull;            /* boolean */
  int eltmark;            /* boolean */
  int lo, hi;             /* k bit word */
  int ind, val;           /* n bit word */
  int mapfull;            /* boolean */
  int mapCode;            /* k bit word */
};

/* The state of the machine consists of the cell states.  In this
   simulator, those states are held in an array s.  (In contrast, the
   abstract tree machine simulator holds the cell states in a tree
   structure, and the generated circuit simulator holds the states in
   flip flops organized as registers.  Those alternative approaches
   are lower level, more accurate, and slower to simulate.) */

static struct cell s [ESF_NCELLS];

static struct esf_result esf_result_buffer;

static int count = 0; /* counts the number of esf operations */
static int initialized = 0;  /* if 0 the system needs to be initialized */

/* ----------------------------------------------------------------------
			 Interface functions
---------------------------------------------------------------------- */


void esf_check (esf_result_ptr p)
{ if (p->esf_return_status == 0)
    return;
  else
    printf ("fatal error\n");  /*terminate program with better message */
}

void esf_print_result (esf_result_ptr p)
{ printf ("esf return_status = %d, return_val = %d\n",
          p->esf_return_status, p->esf_return_result);
}

void esfc_init ()
{ esfp_init (0);
}

int esfc_update (int a, int i, int v)
{ return esf_result (esfp_update (a,i,v)) ;
}

int esfc_lookup (int a, int i)
{ return esf_result (esfp_lookup (a,i));
}

int esf_have_result (esf_result_ptr p)
{ if (p->esf_return_status == 0)
    { return 1; }
  else
    {return 0; }
}

int esf_get_result (esf_result_ptr p)
{ return p->esf_return_result;
}

int esf_result (esf_result_ptr p)
{ if (p->esf_return_status == 0)
    { return p->esf_return_result; }
  else
    { printf ("fatal error\n");  /*terminate program with better message */
      exit(EXIT_FAILURE);
    }
}

/* ----------------------------------------------------------------------
			 Primitive functions
---------------------------------------------------------------------- */

/* esfp_init (a) will initialize the system by putting the right
   values in each of the cell states.  The argument (a) isn't used,
   but might be in the future. */

void esfp_init (int a)
{
  int i;
  count++;
  if (ESF_TRACE) { printf ("%d esfp_init %d\n", count, a); }
  for (i=0; i<ESF_NCELLS; i++)
    { s[i].cellid = i;
      s[i].eltfull = 0;
      s[i].eltmark = 0;
      s[i].lo = 0;
      s[i].hi = 0;
      s[i].ind = 0;
      s[i].val = 0;
      s[i].mapfull= 0;
      s[i].mapCode = 0;
    }
  initialized = 1;
  return;
}

void esfp_dump(int a)
{ int i;
  if (initialized==0) {esfp_init (0);};
  count++;
  for (i=0; i<ESF_NCELLS; i++)
    if (s[i].eltfull || s[i].mapfull)
      { printf ("  %d. ", s[i].cellid);
	if (s[i].eltfull) printf ("*"); else printf (" ");
	if (s[i].eltmark) printf ("@"); else printf (" ");
	if (s[i].eltfull)
	  {printf ("lo=%d, hi=%d, [%d] = %d",
		   s[i].lo, s[i].hi,
		   s[i].ind, s[i].val);}
	else printf ("       ");
	if (s[i].mapfull)
	  {printf (" : array %d -> %d",
		   s[i].cellid, s[i].mapCode);}
	printf ("\n");
      }
  return;
}

/* encode returns -1 if no code is found */
int encode (int a)
{ int i, c;
  count++;
  c = -1;
  if (a == ESF_EMPTY)
    c = 0;   /* code of EMPTY is always 0 */
  else if (s[a].mapfull)
    c = s[a].mapCode ;
  else
    c = -1;
  if (ESF_TRACE) printf ("esf_encode %d returns %d\n", a, c);
  return c;
}

int allocateCell ()
{ int i;

  for (i=0; i<ESF_NCELLS; i++)
    if (! s[i].eltfull)
      { 
	return i;
      }
  printf ("%d allocateCell failed\n", count); /* FIX THIS, RETURN ERROR */
}

esf_result_ptr esfp_update (int a, int ind, int val)
{ int i, c, cnew, newname, nn;
  count++;
  if (ESF_TRACE) {printf("esfp_update(%d,%d,%d)\n", a,ind,val); } ;

  /* Find name and codes */
  c = encode (a);              /* code of array being updated */
  cnew = c+1;                  /* code of new array */
  newname = allocateCell ();    /* name of new array */
  if (ESF_TRACE) {printf("esfp_update c=%d cnew=%d newname=%d\n",
			 c, cnew, newname); } ;

  /* Adjust existing elements and map */
  for (i=0; i<ESF_NCELLS; i++)
    { if (s[i].eltfull)
	{ if (s[i].lo > c)  {s[i].lo++;};
	  if (s[i].hi >= c) {s[i].hi++;};}
      if (s[i].mapfull && s[i].mapCode > c)
	{ s[i].mapCode ++;};
    }

  /* Store new element in allocated cell */
  s[newname].eltfull = 1;
  s[newname].eltmark = 0;
  s[newname].lo = cnew;
  s[newname].hi = cnew;
  s[newname].ind= ind;
  s[newname].val= val;
  s[newname].mapfull = 1;
  s[newname].mapCode= cnew;

  /* return result */
  if (ESF_TRACE) {printf("esfp_update returns %d\n",newname); } ;
  esf_result_buffer.esf_return_status = ESF_OK;
  esf_result_buffer.esf_return_result = newname;
  return &esf_result_buffer;
}

esf_result_ptr esfp_lookup (int a, int ind)
{ int i, c, found, clo,chi, lowinterval, result;
  count++;
  if (ESF_TRACE) {printf ("%d esf_lookup %d %d\n", count, a, ind);} ;

  /* Find code */
  c = encode (a);

  /* Find the element */
  found = 0;
  lowinterval = ESF_NCELLS;
  for (i=0; i<ESF_NCELLS; i++)
    { clo = s[i].lo;
      chi = s[i].hi;
      if (s[i].eltfull && clo <= c && c <= chi &&
	  ind == s[i].ind && chi-clo < lowinterval)
	{ found = 1;
	  lowinterval = chi-clo;
	  result = s[i].val; }
    }
  if (found)
    { esf_result_buffer.esf_return_status = ESF_OK;
      esf_result_buffer.esf_return_result = result;
    }
  else
    { esf_result_buffer.esf_return_status = ESF_LOOKUP_NONE;
      esf_result_buffer.esf_return_result = 0;
    }
  return &esf_result_buffer;
}

/* Set eltmark flag in each cell that contains a potential element of a.
   If there are several cells with the same index, they will all be
   marked; thus a marked cell does not necessarily contain an actual
   element of the array. */

void esfp_markarray (int a)
{ int i, c, found, clo,chi, lowinterval, result;
  count++;
  if (ESF_TRACE) {printf ("%d esf_markarray %d %d\n", count, a);} ;
  /* Find code */
  c = encode (a);
  /* Find the element */
  found = 0;
  lowinterval = ESF_NCELLS;
  for (i=0; i<ESF_NCELLS; i++)
    { clo = s[i].lo;
      chi = s[i].hi;
      if (s[i].eltfull && clo <= c && c <= chi)
	s[i].eltmark = 1;
      else
	s[i].eltmark = 0;
    }
}

esf_result_ptr esfp_maxindex (int a)
{ int i, result;
  if (ESF_TRACE) {printf ("%d esf_maxindex %d\n", a);} ;
  esfp_markarray (a);
  if (ESF_TRACE) esfp_dump(0);
  result = INT_MIN;
  for (i=0; i<ESF_NCELLS; i++)
    { if (s[i].eltmark && s[i].ind > result)
	result = s[i].ind;
    }
  if (ESF_TRACE) {printf ("result = %d\n", result);} ;
  esf_result_buffer.esf_return_status = ESF_OK;
  esf_result_buffer.esf_return_result = result;
  return &esf_result_buffer;
}

int esfc_maxindex (int a)
{ return esf_get_result (esfp_maxindex (a));
}


esf_result_ptr esfp_minindex (int a)
{ int i, result;
  if (ESF_TRACE) {printf ("%d esf_minindex %d\n", a);} ;
  esfp_markarray (a);
  result = INT_MAX;
  for (i=0; i<ESF_NCELLS; i++)
    { if (s[i].eltmark && s[i].ind < result)
	result = s[i].ind;
    }
  esf_result_buffer.esf_return_status = ESF_OK;
  esf_result_buffer.esf_return_result = result;
  return &esf_result_buffer;
}

int esfc_minindex (int a)
{ return esf_get_result (esfp_minindex (a));
}

esf_result_ptr esfp_nextindex (int a, int ind)
{ int i, found, result;
  if (ESF_TRACE) {printf ("%d esf_nextindex %d %d\n", a, ind);} ;
  esfp_markarray (a);
  result = ind;
  for (i=0; i<ESF_NCELLS; i++)
    { if (s[i].eltmark && s[i].ind <= result)
	{ s[i].eltmark = 0; }
    }
  result = INT_MAX;
  for (i=0; i<ESF_NCELLS; i++)
    { if (s[i].eltmark && s[i].ind < result)
	{ result = s[i].ind; }
    }

  esf_result_buffer.esf_return_status = ESF_OK;
  esf_result_buffer.esf_return_result = result;
  return &esf_result_buffer;
}

int esfc_nextindex (int a, int ind)
{ return esf_get_result (esfp_nextindex (a,ind));
}

esf_result_ptr esfp_previndex (int a, int ind)
{ int i, found, result;
  if (ESF_TRACE) {printf ("%esfp_previndex %d %d\n", a, ind);} ;
  esfp_markarray (a);
  if (ESF_TRACE) esfp_dump(0);
  result = ind;
  for (i=0; i<ESF_NCELLS; i++)
    { if (s[i].eltmark && s[i].ind >= result)
	{ s[i].eltmark = 0; }
    }
  if (ESF_TRACE) esfp_dump(0);
  result = INT_MIN;
  for (i=0; i<ESF_NCELLS; i++)
    { if (s[i].eltmark && s[i].ind > result)
	{ result = s[i].ind; }
    }

  if (ESF_TRACE) {printf ("esfp_previndex result = %d\n", result);} ;
  esf_result_buffer.esf_return_status = ESF_OK;
  esf_result_buffer.esf_return_result = result;
  return &esf_result_buffer;
}

int esfc_previndex (int a, int ind)
{ return esf_get_result (esfp_previndex (a,ind));
}
